"""Test translation from flat dict configs to KvConfig layer stacks."""

from copy import deepcopy
from typing import Any, Iterator, cast, get_args

import msgspec

import pytest

from lmstudio import BaseModel, DictObject, LMStudioValueError
from lmstudio.schemas import LMStudioStruct
from lmstudio._kv_config import (
    ToServerKeymap,
    TO_SERVER_LOAD_EMBEDDING,
    TO_SERVER_LOAD_LLM,
    TO_SERVER_PREDICTION,
    load_config_to_kv_config_stack,
    parse_server_config,
    prediction_config_to_kv_config_stack,
)
from lmstudio._sdk_models import (
    EmbeddingLoadModelConfig,
    EmbeddingLoadModelConfigDict,
    GpuSetting,
    GpuSettingDict,
    GpuSplitConfigDict,
    KvConfigFieldDict,
    KvConfigStackDict,
    LlmLoadModelConfig,
    LlmLoadModelConfigDict,
    LlmPredictionConfig,
    LlmPredictionConfigDict,
    LlmSplitStrategy,
)

# Note: configurations below are just for data manipulation round-trip testing,
#       so they don't necessarily make sense as actual model configurations

GPU_CONFIG: GpuSettingDict = {
    "mainGpu": 0,
    "ratio": 0.5,
    "splitStrategy": "evenly",
    "disabledGpus": [1, 2],
}
SC_GPU_CONFIG = {
    "main_gpu": 0,
    "ratio": 0.5,
    "split_strategy": "evenly",
    "disabled_gpus": [1, 2],
}

LOAD_CONFIG_EMBEDDING: EmbeddingLoadModelConfigDict = {
    "contextLength": 1978,
    "gpu": GPU_CONFIG,
    "keepModelInMemory": True,
    "ropeFrequencyBase": 10.0,
    "ropeFrequencyScale": 1.5,
    "tryMmap": False,
}

SC_LOAD_CONFIG_EMBEDDING = {
    "context_length": 1978,
    "gpu": SC_GPU_CONFIG,
    "keep_model_in_memory": True,
    "rope_frequency_base": 10.0,
    "rope_frequency_scale": 1.5,
    "try_mmap": False,
}

LOAD_CONFIG_LLM: LlmLoadModelConfigDict = {
    "contextLength": 1978,
    "evalBatchSize": 42,
    "flashAttention": False,
    "gpu": GPU_CONFIG,
    "gpuStrictVramCap": False,
    "keepModelInMemory": True,
    "llamaKCacheQuantizationType": "q8_0",
    "llamaVCacheQuantizationType": "f32",
    "numExperts": 0,
    "offloadKVCacheToGpu": False,
    "ropeFrequencyBase": 10.0,
    "ropeFrequencyScale": 1.5,
    "seed": 313,
    "tryMmap": False,
    "useFp16ForKVCache": True,
}

SC_LOAD_CONFIG_LLM = {
    "context_length": 1978,
    "eval_batch_size": 42,
    "flash_attention": False,
    "gpu": SC_GPU_CONFIG,
    "gpu_strict_vram_cap": False,
    "keep_model_in_memory": True,
    "llama_k_cache_quantization_type": "q8_0",
    "llama_v_cache_quantization_type": "f32",
    "num_experts": 0,
    "offload_kv_cache_to_gpu": False,
    "rope_frequency_base": 10.0,
    "rope_frequency_scale": 1.5,
    "seed": 313,
    "try_mmap": False,
    "use_fp16_for_kv_cache": True,
}

PREDICTION_CONFIG: LlmPredictionConfigDict = {
    "contextOverflowPolicy": "rollingWindow",
    "cpuThreads": 7,
    "draftModel": "some-model-key",
    "maxTokens": 1234,
    "minPSampling": 5.5,
    "promptTemplate": {
        "type": "manual",
        "stopStrings": ["Nevermore"],
        "manualPromptTemplate": {
            "beforeSystem": "example prefix",
            "afterSystem": "example suffix",
            "beforeUser": "example prefix",
            "afterUser": "example suffix",
            "beforeAssistant": "example prefix",
            "afterAssistant": "example suffix",
        },
    },
    "reasoningParsing": {"enabled": False, "startString": "", "endString": ""},
    "repeatPenalty": 6.5,
    "speculativeDecodingNumDraftTokensExact": 2,
    "speculativeDecodingMinDraftLengthToConsider": 5,
    "speculativeDecodingMinContinueDraftingProbability": 0.1,
    "stopStrings": ["Banana!"],
    "structured": {"type": "json", "jsonSchema": {"type": "string"}},
    "temperature": 2.5,
    "toolCallStopStrings": ["yellow"],
    "rawTools": {"type": "none"},
    "topKSampling": 3.5,
    "topPSampling": 4.5,
}

SC_PREDICTION_CONFIG = {
    "context_overflow_policy": "rollingWindow",
    "cpu_threads": 7,
    "draft_model": "some-model-key",
    "max_tokens": 1234,
    "min_p_sampling": 5.5,
    "prompt_template": {
        "type": "manual",
        "stop_strings": ["Nevermore"],
        "manual_prompt_template": {
            "before_system": "example prefix",
            "after_system": "example suffix",
            "before_user": "example prefix",
            "after_user": "example suffix",
            "before_assistant": "example prefix",
            "after_assistant": "example suffix",
        },
    },
    "reasoning_parsing": {"enabled": False, "start_string": "", "end_string": ""},
    "repeat_penalty": 6.5,
    "speculative_decoding_num_draft_tokens_exact": 2,
    "speculative_decoding_min_draft_length_to_consider": 5,
    "speculative_decoding_min_continue_drafting_probability": 0.1,
    "stop_strings": ["Banana!"],
    "structured": {"type": "json", "json_schema": {"type": "string"}},
    "temperature": 2.5,
    "tool_call_stop_strings": ["yellow"],
    "raw_tools": {"type": "none"},
    "top_k_sampling": 3.5,
    "top_p_sampling": 4.5,
}


CONFIG_DICTS = (
    GPU_CONFIG,
    LOAD_CONFIG_EMBEDDING,
    LOAD_CONFIG_LLM,
    PREDICTION_CONFIG,
)

SC_DICTS = (
    SC_GPU_CONFIG,
    SC_LOAD_CONFIG_EMBEDDING,
    SC_LOAD_CONFIG_LLM,
    SC_PREDICTION_CONFIG,
)

CONFIG_TYPES = (
    GpuSetting,
    EmbeddingLoadModelConfig,
    LlmLoadModelConfig,
    LlmPredictionConfig,
)

KEYMAP_DICTS = (
    TO_SERVER_LOAD_EMBEDDING,
    TO_SERVER_LOAD_LLM,
    TO_SERVER_PREDICTION,
)

KEYMAP_TYPES = CONFIG_TYPES[1:]


# Define strict variants that don't implicitly discard unknown keys
class GpuSettingStrict(GpuSetting, forbid_unknown_fields=True):
    pass


class EmbeddingLoadModelConfigStrict(
    EmbeddingLoadModelConfig, forbid_unknown_fields=True
):
    pass


class LlmLoadModelConfigStrict(LlmLoadModelConfig, forbid_unknown_fields=True):
    pass


class LlmPredictionConfigStrict(LlmPredictionConfig, forbid_unknown_fields=True):
    pass


STRICT_TYPES = (
    GpuSettingStrict,
    EmbeddingLoadModelConfigStrict,
    LlmLoadModelConfigStrict,
    LlmPredictionConfigStrict,
)

# The "raw" debugging field is a special case, with TBD handling
_NOT_YET_MAPPED = {"raw"}


@pytest.mark.parametrize("config_dict,config_type", zip(CONFIG_DICTS, CONFIG_TYPES))
def test_struct_field_coverage(
    config_dict: DictObject, config_type: LMStudioStruct[Any]
) -> None:
    # Ensure all expected keys are covered (even those with default values)
    mapped_keys = set(config_type.__struct_encode_fields__)
    expected_keys = config_dict.keys()
    missing_keys = expected_keys - mapped_keys
    assert not missing_keys
    # Ensure no extra keys are mistakenly defined
    unknown_keys = mapped_keys - expected_keys - _NOT_YET_MAPPED
    assert not unknown_keys
    # Ensure the config can be loaded
    config_struct = config_type._from_api_dict(config_dict)
    assert config_struct.to_dict() == config_dict


@pytest.mark.parametrize(
    "input_dict,expected_dict,config_type", zip(SC_DICTS, CONFIG_DICTS, STRICT_TYPES)
)
def test_snake_case_conversion(
    input_dict: DictObject, expected_dict: DictObject, config_type: LMStudioStruct[Any]
) -> None:
    # Ensure snake case keys are converted to camelCase for user-supplied dicts
    config_struct = config_type.from_dict(input_dict)
    assert config_struct.to_dict() == expected_dict
    # Ensure no conversion is applied when reading API dicts
    with pytest.raises(msgspec.ValidationError):
        config_type._from_api_dict(input_dict)


@pytest.mark.parametrize("keymap,config_type", zip(KEYMAP_DICTS, KEYMAP_TYPES))
def test_kv_stack_field_coverage(
    keymap: ToServerKeymap, config_type: LMStudioStruct[Any]
) -> None:
    # Ensure all expected keys are covered (even those with default values)
    mapped_keys = keymap.keys()
    expected_keys = set(config_type.__struct_encode_fields__)
    missing_keys = expected_keys - mapped_keys - _NOT_YET_MAPPED
    assert not missing_keys
    # Ensure no extra keys are mistakenly defined
    unknown_keys = mapped_keys - expected_keys
    assert not unknown_keys


EXPECTED_KV_STACK_LOAD_EMBEDDING: KvConfigStackDict = {
    "layers": [
        {
            "config": {
                "fields": [
                    {"key": "embedding.load.contextLength", "value": 1978},
                    {
                        "key": "embedding.load.llama.acceleration.offloadRatio",
                        "value": 0.5,
                    },
                    {
                        "key": "load.gpuSplitConfig",
                        "value": {
                            "disabledGpus": [1, 2],
                            "strategy": "evenly",
                            "priority": [],
                            "customRatio": [],
                        },
                    },
                    {"key": "embedding.load.llama.keepModelInMemory", "value": True},
                    {
                        "key": "embedding.load.llama.ropeFrequencyBase",
                        "value": {"checked": True, "value": 10.0},
                    },
                    {
                        "key": "embedding.load.llama.ropeFrequencyScale",
                        "value": {"checked": True, "value": 1.5},
                    },
                    {"key": "embedding.load.llama.tryMmap", "value": False},
                ],
            },
            "layerName": "apiOverride",
        },
    ],
}

EXPECTED_KV_STACK_LOAD_LLM: KvConfigStackDict = {
    "layers": [
        {
            "layerName": "apiOverride",
            "config": {
                "fields": [
                    {"key": "llm.load.contextLength", "value": 1978},
                    {"key": "llm.load.llama.acceleration.offloadRatio", "value": 0.5},
                    {
                        "key": "load.gpuSplitConfig",
                        "value": {
                            "disabledGpus": [1, 2],
                            "strategy": "evenly",
                            "priority": [],
                            "customRatio": [],
                        },
                    },
                    {"key": "llm.load.llama.evalBatchSize", "value": 42},
                    {"key": "llm.load.llama.flashAttention", "value": False},
                    {
                        "key": "llm.load.llama.kCacheQuantizationType",
                        "value": {"checked": True, "value": "q8_0"},
                    },
                    {"key": "llm.load.llama.keepModelInMemory", "value": True},
                    {
                        "key": "llm.load.llama.ropeFrequencyBase",
                        "value": {"checked": True, "value": 10.0},
                    },
                    {
                        "key": "llm.load.llama.ropeFrequencyScale",
                        "value": {"checked": True, "value": 1.5},
                    },
                    {"key": "llm.load.llama.tryMmap", "value": False},
                    {"key": "llm.load.llama.useFp16ForKVCache", "value": True},
                    {
                        "key": "llm.load.llama.vCacheQuantizationType",
                        "value": {"checked": True, "value": "f32"},
                    },
                    {"key": "llm.load.numExperts", "value": 0},
                    {"key": "llm.load.offloadKVCacheToGpu", "value": False},
                    {"key": "llm.load.seed", "value": {"checked": True, "value": 313}},
                    {"key": "load.gpuStrictVramCap", "value": False},
                ]
            },
        }
    ]
}

EXPECTED_KV_STACK_PREDICTION: KvConfigStackDict = {
    "layers": [
        {
            "config": {
                "fields": [
                    {
                        "key": "llm.prediction.contextOverflowPolicy",
                        "value": "rollingWindow",
                    },
                    {"key": "llm.prediction.llama.cpuThreads", "value": 7},
                    {
                        "key": "llm.prediction.maxPredictedTokens",
                        "value": {"checked": True, "value": 1234},
                    },
                    {
                        "key": "llm.prediction.minPSampling",
                        "value": {"checked": True, "value": 5.5},
                    },
                    {
                        "key": "llm.prediction.promptTemplate",
                        "value": {
                            "manualPromptTemplate": {
                                "afterAssistant": "example suffix",
                                "afterSystem": "example suffix",
                                "afterUser": "example suffix",
                                "beforeAssistant": "example prefix",
                                "beforeSystem": "example prefix",
                                "beforeUser": "example prefix",
                            },
                            "stopStrings": ["Nevermore"],
                            "type": "manual",
                        },
                    },
                    {
                        "key": "llm.prediction.reasoning.parsing",
                        "value": {"enabled": False, "startString": "", "endString": ""},
                    },
                    {
                        "key": "llm.prediction.repeatPenalty",
                        "value": {"checked": True, "value": 6.5},
                    },
                    {
                        "key": "llm.prediction.speculativeDecoding.draftModel",
                        "value": "some-model-key",
                    },
                    {
                        "key": "llm.prediction.speculativeDecoding.minContinueDraftingProbability",
                        "value": 0.1,
                    },
                    {
                        "key": "llm.prediction.speculativeDecoding.minDraftLengthToConsider",
                        "value": 5,
                    },
                    {
                        "key": "llm.prediction.speculativeDecoding.numDraftTokensExact",
                        "value": 2,
                    },
                    {"key": "llm.prediction.stopStrings", "value": ["Banana!"]},
                    {
                        "key": "llm.prediction.structured",
                        "value": {"jsonSchema": {"type": "string"}, "type": "json"},
                    },
                    {"key": "llm.prediction.temperature", "value": 2.5},
                    {
                        "key": "llm.prediction.toolCallStopStrings",
                        "value": ["yellow"],
                    },
                    {"key": "llm.prediction.tools", "value": {"type": "none"}},
                    {"key": "llm.prediction.topKSampling", "value": 3.5},
                    {
                        "key": "llm.prediction.topPSampling",
                        "value": {"checked": True, "value": 4.5},
                    },
                ],
            },
            "layerName": "apiOverride",
        },
    ],
}


@pytest.mark.parametrize(
    "config_dict", (LOAD_CONFIG_EMBEDDING, SC_LOAD_CONFIG_EMBEDDING)
)
def test_kv_stack_load_config_embedding(config_dict: DictObject) -> None:
    kv_stack = load_config_to_kv_config_stack(config_dict, EmbeddingLoadModelConfig)
    assert kv_stack.to_dict() == EXPECTED_KV_STACK_LOAD_EMBEDDING


@pytest.mark.parametrize("config_dict", (LOAD_CONFIG_LLM, SC_LOAD_CONFIG_LLM))
def test_kv_stack_load_config_llm(config_dict: DictObject) -> None:
    kv_stack = load_config_to_kv_config_stack(config_dict, LlmLoadModelConfig)
    assert kv_stack.to_dict() == EXPECTED_KV_STACK_LOAD_LLM


def test_parse_server_config_load_embedding() -> None:
    server_config = EXPECTED_KV_STACK_LOAD_EMBEDDING["layers"][0]["config"]
    expected_client_config = deepcopy(LOAD_CONFIG_EMBEDDING)
    gpu_settings_dict = expected_client_config["gpu"]
    assert gpu_settings_dict is not None
    del gpu_settings_dict["mainGpu"]  # This is not reported with "evenly" strategy
    assert parse_server_config(server_config) == expected_client_config


def test_parse_server_config_load_llm() -> None:
    server_config = EXPECTED_KV_STACK_LOAD_LLM["layers"][0]["config"]
    expected_client_config = deepcopy(LOAD_CONFIG_LLM)
    gpu_settings_dict = expected_client_config["gpu"]
    assert gpu_settings_dict is not None
    del gpu_settings_dict["mainGpu"]  # This is not reported with "evenly" strategy
    assert parse_server_config(server_config) == expected_client_config


def _gpu_split_strategies() -> Iterator[LlmSplitStrategy]:
    # Ensure all GPU split strategies are checked (these aren't simple structural transforms,
    # so the default test case doesn't provide adequate test coverage )
    for split_strategy in get_args(LlmSplitStrategy):
        yield split_strategy


def _find_config_field(
    stack_dict: KvConfigStackDict, key: str
) -> tuple[int, KvConfigFieldDict]:
    for enumerated_field in enumerate(stack_dict["layers"][0]["config"]["fields"]):
        if enumerated_field[1]["key"] == key:
            return enumerated_field
    raise KeyError(key)


def _del_config_field(stack_dict: KvConfigStackDict, key: str) -> None:
    field_index = _find_config_field(stack_dict, key)[0]
    field_list = cast(list[Any], stack_dict["layers"][0]["config"]["fields"])
    del field_list[field_index]


def _find_config_value(stack_dict: KvConfigStackDict, key: str) -> Any:
    return _find_config_field(stack_dict, key)[1]["value"]


def _append_invalid_config_field(stack_dict: KvConfigStackDict, key: str) -> None:
    field_list = cast(list[Any], stack_dict["layers"][0]["config"]["fields"])
    field_list.append({"key": key})


@pytest.mark.parametrize("split_strategy", _gpu_split_strategies())
def test_gpu_split_strategy_config(split_strategy: LlmSplitStrategy) -> None:
    # GPU config mapping is complex enough to need some additional testing
    input_camelCase = deepcopy(LOAD_CONFIG_LLM)
    input_snake_case = deepcopy(SC_LOAD_CONFIG_LLM)
    gpu_camelCase: GpuSettingDict = cast(Any, input_camelCase["gpu"])
    gpu_snake_case: dict[str, Any] = cast(Any, input_snake_case["gpu"])
    expected_stack = deepcopy(EXPECTED_KV_STACK_LOAD_LLM)
    expected_server_config = expected_stack["layers"][0]["config"]
    gpu_camelCase["splitStrategy"] = gpu_snake_case["split_strategy"] = split_strategy
    if split_strategy == GPU_CONFIG["splitStrategy"]:
        assert split_strategy == "evenly", (
            "Unexpected default LLM GPU offload split strategy (missing test case update?)"
        )
        # There is no main GPU when the split strategy is even across GPUs
        del gpu_camelCase["mainGpu"]
        del gpu_snake_case["main_gpu"]
    elif split_strategy == "favorMainGpu":
        expected_split_config: GpuSplitConfigDict = _find_config_value(
            expected_stack, "load.gpuSplitConfig"
        )
        expected_split_config["strategy"] = "priorityOrder"
        main_gpu = GPU_CONFIG["mainGpu"]
        assert main_gpu is not None
        expected_split_config["priority"] = [main_gpu]
    else:
        assert split_strategy is None, (
            "Unknown LLM GPU offload split strategy (missing test case update?)"
        )
    # Check given GPU config maps as expected in both directions
    kv_stack = load_config_to_kv_config_stack(input_camelCase, LlmLoadModelConfig)
    assert kv_stack.to_dict() == expected_stack
    kv_stack = load_config_to_kv_config_stack(input_snake_case, LlmLoadModelConfig)
    assert kv_stack.to_dict() == expected_stack
    assert parse_server_config(expected_server_config) == input_camelCase
    # Check a malformed ratio field is tolerated
    gpu_camelCase["ratio"] = gpu_snake_case["ratio"] = None
    _del_config_field(expected_stack, "llm.load.llama.acceleration.offloadRatio")
    _append_invalid_config_field(
        expected_stack, "llm.load.llama.acceleration.offloadRatio"
    )
    assert parse_server_config(expected_server_config) == input_camelCase
    # Check mapping works if no explicit offload ratio is specified
    _del_config_field(expected_stack, "llm.load.llama.acceleration.offloadRatio")
    kv_stack = load_config_to_kv_config_stack(input_camelCase, LlmLoadModelConfig)
    assert kv_stack.to_dict() == expected_stack
    kv_stack = load_config_to_kv_config_stack(input_snake_case, LlmLoadModelConfig)
    assert kv_stack.to_dict() == expected_stack
    del gpu_camelCase["ratio"]
    assert parse_server_config(expected_server_config) == input_camelCase


@pytest.mark.parametrize("config_dict", (PREDICTION_CONFIG, SC_PREDICTION_CONFIG))
def test_kv_stack_prediction_config(config_dict: DictObject) -> None:
    # MyPy complains here that it can't be sure the dict has all the right keys
    # It is correct about that, but we want to ensure it is handled at runtime
    structured, kv_stack = prediction_config_to_kv_config_stack(None, config_dict)  # type: ignore[arg-type]
    assert structured
    assert kv_stack.to_dict() == EXPECTED_KV_STACK_PREDICTION


def test_kv_stack_prediction_config_conflict() -> None:
    with pytest.raises(
        LMStudioValueError, match="Cannot specify.*response_format.*structured"
    ):
        prediction_config_to_kv_config_stack(BaseModel, PREDICTION_CONFIG)


# TODO: Come up with a way to do the strict checks that applies to nested dicts
#       (this will most likely involve changing the data model code generation)
# def test_nested_unknown_keys() -> None:
#     config = LOAD_CONFIG_EMBEDDING.copy()
#     LOAD_CONFIG_EMBEDDING["gpu"] = SC_GPU_CONFIG
#     with pytest.raises(msgspec.ValidationError):
#         EmbeddingLoadModelConfigStrict._from_api_dict(config)



"""Test making simple predictions with the API."""

import asyncio
import logging

from typing import Any

import pytest
from pytest import LogCaptureFixture as LogCap
from pytest_subtests import SubTests

from lmstudio import (
    AsyncClient,
    Chat,
    LlmPredictionConfig,
    LMStudioValueError,
    PredictionResult,
    ToolFunctionDef,
)
from lmstudio.json_api import ChatResponseEndpoint
from lmstudio._sdk_models import LlmToolParameters

from .support import (
    ADDITION_TOOL_SPEC,
    EXPECTED_LLM_ID,
    MAX_PREDICTED_TOKENS,
    SHORT_PREDICTION_CONFIG,
    log_adding_two_integers,
)

SC_PREDICTION_CONFIG = {
    "max_tokens": MAX_PREDICTED_TOKENS,
    "temperature": 0,
}


def test_prediction_config_translation() -> None:
    # Ensure prediction config with snake_case keys is translated
    assert SC_PREDICTION_CONFIG != SHORT_PREDICTION_CONFIG
    # We expect this to fail static type checking
    struct_config = LlmPredictionConfig.from_dict(SC_PREDICTION_CONFIG)  # type: ignore[arg-type]
    expected_struct_config = LlmPredictionConfig(
        max_tokens=MAX_PREDICTED_TOKENS, temperature=0
    )
    assert struct_config == expected_struct_config
    assert struct_config.to_dict() == SHORT_PREDICTION_CONFIG


@pytest.mark.asyncio
@pytest.mark.slow
@pytest.mark.lmstudio
async def test_concurrent_predictions(caplog: LogCap, subtests: SubTests) -> None:
    phrase = "I am a robot"
    refusal = "I cannot repeat"
    request = f"Repeat the phrase '{phrase}' ten times."
    history = Chat(
        "You are a very obedient automatic assistant that always behaves as requested. "
        "You give terse responses."
    )
    history.add_user_message(request)

    caplog.set_level(logging.DEBUG)
    model_id = EXPECTED_LLM_ID
    async with AsyncClient() as client:
        session = client.llm

        async def _request_response() -> PredictionResult:
            llm = await session.model(model_id)
            return await llm.respond(
                history=history,
                config=SHORT_PREDICTION_CONFIG,
            )

        # Note: too many parallel requests risks overloading the local API server
        requests = [_request_response() for i in range(5)]
        responses = await asyncio.gather(*requests)
    subtests_passed = 0
    for i, response in enumerate(responses):
        with subtests.test("Check prediction response", i=i):
            # The responses are variable enough that not much can be checked here
            assert phrase in response.content or refusal in response.content
            subtests_passed += 1

    # Work around pytest-subtests not showing full output when subtests fail
    # https://github.com/pytest-dev/pytest-subtests/issues/76
    assert subtests_passed == len(responses), "Fail due to failed subtest(s)"


# TODO: write sync concurrent predictions test with external locking and concurrent.futures


def test_tool_def_from_callable() -> None:
    default_def = ToolFunctionDef.from_callable(log_adding_two_integers)
    assert default_def == ToolFunctionDef(
        name=log_adding_two_integers.__name__,
        description="Log adding two integers together.",
        parameters=ADDITION_TOOL_SPEC["parameters"],
        implementation=log_adding_two_integers,
    )
    custom_def = ToolFunctionDef.from_callable(
        log_adding_two_integers, name="add", description="Add two numbers"
    )
    assert custom_def == ToolFunctionDef(**ADDITION_TOOL_SPEC)


def test_parse_tools() -> None:
    addition_def = ToolFunctionDef.from_callable(
        log_adding_two_integers, name="add_as_tool_def"
    )
    tools: list[Any] = [ADDITION_TOOL_SPEC, addition_def, log_adding_two_integers]
    expected_implementations = {
        "add": log_adding_two_integers,
        "add_as_tool_def": log_adding_two_integers,
        "log_adding_two_integers": log_adding_two_integers,
    }
    expected_names = list(expected_implementations.keys())
    expected_param_schemas = 3 * [
        LlmToolParameters(
            type="object",
            properties={"a": {"type": "integer"}, "b": {"type": "integer"}},
            required=["a", "b"],
            additional_properties=None,
        )
    ]
    llm_tools, client_map = ChatResponseEndpoint.parse_tools(tools)
    assert llm_tools.tools is not None
    assert [t.function.name for t in llm_tools.tools] == expected_names
    assert [t.function.parameters for t in llm_tools.tools] == expected_param_schemas
    assert client_map.keys() == set(expected_names)
    client_tools = {k: v[1] for k, v in client_map.items()}
    assert client_tools == expected_implementations


def test_duplicate_tool_names_rejected() -> None:
    addition_def = ToolFunctionDef.from_callable(log_adding_two_integers, name="add")
    tools: list[Any] = [ADDITION_TOOL_SPEC, addition_def]
    with pytest.raises(
        LMStudioValueError, match="Duplicate tool names are not permitted"
    ):
        ChatResponseEndpoint.parse_tools(tools)


async def example_async_tool() -> int:
    """Example asynchronous tool definition"""
    return 42


def test_async_tool_rejected() -> None:
    tools: list[Any] = [example_async_tool]
    with pytest.raises(LMStudioValueError, match=".*example_async_tool.*not supported"):
        ChatResponseEndpoint.parse_tools(tools)


def test_async_tool_accepted() -> None:
    tools: list[Any] = [example_async_tool]
    llm_tools, client_map = ChatResponseEndpoint.parse_tools(tools, allow_async=True)
    assert llm_tools.tools is not None
    assert len(llm_tools.tools) == 1
    assert len(client_map) == 1



"""Invoking and supporting tools provider hook implementations."""

import asyncio

from contextvars import ContextVar
from dataclasses import dataclass
from traceback import format_tb
from typing import Any, Awaitable, Callable, Generic, Iterable, TypeAlias, TypeVar
from typing_extensions import (
    # Native in 3.11+
    assert_never,
)

from anyio import create_task_group
from anyio.abc import TaskGroup
from msgspec import convert, to_builtins

from ..._logging import new_logger, LogEventContext
from ...schemas import DictObject, EmptyDict
from ...json_api import (
    ChannelCommonRxEvent,
    ChannelEndpoint,
    ChannelFinishedEvent,
    ChannelRxEvent,
    ChatResponseEndpoint,
    ClientToolMap,
    SendMessageAsync,
    ToolDefinition,
)
from ..._sdk_models import (
    PluginToolCallComplete,
    PluginToolCallCompleteDict,
    PluginToolCallErrorDict,
    PluginToolCallStatusDict,
    PluginToolCallWarnDict,
    ProvideToolsInitSession,
    ProvideToolsAbortCall,
    ProvideToolsCallTool,
    ProvideToolsInitFailedDict,
    ProvideToolsInitializedDict,
    SerializedLMSExtendedErrorDict,
)

from ..config_schemas import BaseConfigSchema
from ..sdk_api import LMStudioPluginRuntimeError
from .common import (
    _AsyncSessionPlugins,
    HookController,
    ServerRequestError,
    TPluginConfigSchema,
    TGlobalConfigSchema,
)

# Available as lmstudio.plugin.hooks.*
__all__ = [
    "AsyncToolCallContext",
    "ToolCallContext",
    "ToolsProviderController",
    "ToolsProviderHook",
    "run_tools_provider",
    "get_tool_call_context",
    "get_tool_call_context_async",
]


class ProvideToolsDiscardSessionEvent(ChannelRxEvent[str]):
    pass


class ProvideToolsInitSessionEvent(ChannelRxEvent[ProvideToolsInitSession]):
    pass


class ProvideToolsCallToolEvent(ChannelRxEvent[ProvideToolsCallTool]):
    pass


class ProvideToolsAbortCallEvent(ChannelRxEvent[ProvideToolsAbortCall]):
    pass


PromptPreprocessingRxEvent: TypeAlias = (
    ProvideToolsDiscardSessionEvent
    | ProvideToolsInitSessionEvent
    | ProvideToolsCallToolEvent
    | ProvideToolsAbortCallEvent
    | ChannelCommonRxEvent
)


class ToolsProviderEndpoint(
    ChannelEndpoint[tuple[str, str], PromptPreprocessingRxEvent, EmptyDict]
):
    """API channel endpoint to accept prompt preprocessing requests."""

    _API_ENDPOINT = "setToolsProvider"
    _NOTICE_PREFIX = "Providing tools"

    def __init__(self) -> None:
        super().__init__({})

    def iter_message_events(
        self, contents: DictObject | None
    ) -> Iterable[PromptPreprocessingRxEvent]:
        match contents:
            case None:
                # Server can only terminate the link by closing the websocket
                pass
            case {"type": "discardSession", "sessionId": str(session_id)}:
                yield ProvideToolsDiscardSessionEvent(session_id)
            case {"type": "initSession"} as init_session_dict:
                init_session = ProvideToolsInitSession._from_any_api_dict(
                    init_session_dict
                )
                yield ProvideToolsInitSessionEvent(init_session)
            case {"type": "callTool"} as tool_call_dict:
                tool_call = ProvideToolsCallTool._from_any_api_dict(tool_call_dict)
                yield ProvideToolsCallToolEvent(tool_call)
            case {"type": "abortToolCall"} as abort_tool_call_dict:
                abort_tool_call = ProvideToolsAbortCall._from_any_api_dict(
                    abort_tool_call_dict
                )
                yield ProvideToolsAbortCallEvent(abort_tool_call)
            case unmatched:
                self.report_unknown_message(unmatched)

    def handle_rx_event(self, event: PromptPreprocessingRxEvent) -> None:
        match event:
            case ProvideToolsDiscardSessionEvent(session_id):
                self._logger.debug(f"Terminating {session_id}", session_id=session_id)
            case ProvideToolsInitSessionEvent(request):
                self._logger.debug(
                    "Received tools session request", session_id=request.session_id
                )
            case ProvideToolsCallToolEvent(request):
                self._logger.debug(
                    "Received tool call request",
                    session_id=request.session_id,
                    call_id=request.call_id,
                )
            case ProvideToolsAbortCallEvent(request):
                self._logger.debug(
                    "Received tool abort request",
                    session_id=request.session_id,
                    call_id=request.call_id,
                )
            case ChannelFinishedEvent(_):
                pass
            case _:
                assert_never(event)


class ToolsProviderController(
    HookController[ProvideToolsInitSession, TPluginConfigSchema, TGlobalConfigSchema]
):
    """API access for tools provider hook implementations."""

    def __init__(
        self,
        session: _AsyncSessionPlugins,
        request: ProvideToolsInitSession,
        plugin_config_schema: type[TPluginConfigSchema],
        global_config_schema: type[TGlobalConfigSchema],
    ) -> None:
        """Initialize prompt preprocessor hook controller."""
        super().__init__(session, request, plugin_config_schema, global_config_schema)
        self.session_id = request.session_id


ToolsProviderHook = Callable[
    [ToolsProviderController[Any, Any]], Awaitable[Iterable[ToolDefinition]]
]

T = TypeVar("T")


class _BaseToolCallContext:
    """API access to update a tool call UI status block in-place."""

    def __init__(
        self,
        session_id: str,
        call_id: str,
        send_json: SendMessageAsync,
    ) -> None:
        """Initialize status block controller."""
        self.session_id = session_id
        self.call_id = call_id
        self._send_json = send_json

    def _make_status(self, message: str) -> PluginToolCallStatusDict:
        return PluginToolCallStatusDict(
            type="toolCallStatus",
            sessionId=self.session_id,
            callId=self.call_id,
            statusText=message,
        )

    def _make_warning(self, message: str) -> PluginToolCallWarnDict:
        return PluginToolCallWarnDict(
            type="toolCallWarn",
            sessionId=self.session_id,
            callId=self.call_id,
            warnText=message,
        )


class AsyncToolCallContext(_BaseToolCallContext):
    """Asynchronous API access to update a tool call UI status block in-place."""

    async def notify_status(self, message: str) -> None:
        """Report tool progress update in the task status block."""
        await self._send_json(self._make_status(message))

    async def notify_warning(self, message: str) -> None:
        """Report tool warning in the task status block."""
        await self._send_json(self._make_warning(message))


class ToolCallContext(_BaseToolCallContext):
    """Synchronous API access to update a tool call UI status block in-place."""

    def __init__(
        self,
        session_id: str,
        call_id: str,
        send_json: SendMessageAsync,
    ) -> None:
        """Initialize synchronous status block controller."""
        super().__init__(session_id, call_id, send_json)
        # Sync call context is created in the plugin's async comms loop
        self._loop = asyncio.get_running_loop()

    def _send_json_sync(self, data: DictObject) -> None:
        future = asyncio.run_coroutine_threadsafe(self._send_json(data), self._loop)
        future.result()

    def notify_status(self, message: str) -> None:
        """Report tool progress update in the task status block."""
        self._send_json_sync(self._make_status(message))

    def notify_warning(self, message: str) -> None:
        """Report tool warning in the task status block."""
        self._send_json_sync(self._make_warning(message))


_LMS_TOOL_CALL_SYNC: ContextVar[ToolCallContext] = ContextVar("_LMS_TOOL_CALL_SYNC")
_LMS_TOOL_CALL_ASYNC: ContextVar[AsyncToolCallContext] = ContextVar(
    "_LMS_TOOL_CALL_ASYNC"
)


def get_tool_call_context() -> ToolCallContext:
    """Get synchronous tool call context."""
    if _LMS_TOOL_CALL_ASYNC.get(None) is not None:
        msg = "Use 'get_tool_call_context_async()' in asynchronous tool definition"
        raise LMStudioPluginRuntimeError(msg)
    return _LMS_TOOL_CALL_SYNC.get()


def get_tool_call_context_async() -> AsyncToolCallContext:
    """Get asynchronous tool call context."""
    if _LMS_TOOL_CALL_SYNC.get(None) is not None:
        msg = "Use 'get_tool_call_context()' in synchronous tool definition"
        raise LMStudioPluginRuntimeError(msg)
    return _LMS_TOOL_CALL_ASYNC.get()


class ToolCallHandler:
    def __init__(
        self,
        plugin_name: str,
        session_id: str,
        provided_tools: ClientToolMap,
        log_context: LogEventContext,
    ) -> None:
        self.plugin_name = plugin_name
        self.session_id = session_id
        self._provided_tools = provided_tools
        self._queue: asyncio.Queue[ProvideToolsCallTool | None] = asyncio.Queue()
        self._abort_events: dict[str, asyncio.Event] = {}
        self._logger = logger = new_logger(__name__)
        logger.update_context(log_context, session_id=session_id)

    async def _cancel_on_event(
        self, tg: TaskGroup, event: asyncio.Event, message: str
    ) -> None:
        await event.wait()
        self._logger.info(message)
        tg.cancel_scope.cancel()

    async def start_tool_call(self, tool_call: ProvideToolsCallTool) -> None:
        await self._queue.put(tool_call)

    # TODO: Reduce code duplication with the ChatResponseEndpoint definition
    async def _call_async_tool(
        self,
        call_id: str,
        implementation: Callable[..., Awaitable[Any]],
        kwds: DictObject,
        send_json: SendMessageAsync,
    ) -> PluginToolCallCompleteDict:
        assert _LMS_TOOL_CALL_SYNC.get(None) is None
        call_context = AsyncToolCallContext(self.session_id, call_id, send_json)
        _LMS_TOOL_CALL_ASYNC.set(call_context)
        call_result = await implementation(**kwds)
        return PluginToolCallComplete(
            session_id=self.session_id,
            call_id=call_id,
            result=call_result,
        ).to_dict()

    def _call_sync_tool(
        self,
        call_id: str,
        implementation: Callable[..., Any],
        kwds: DictObject,
        send_json: SendMessageAsync,
    ) -> Awaitable[PluginToolCallCompleteDict]:
        # Ensure synchronous tools can't block the plugin's async comms thread
        call_context = ToolCallContext(self.session_id, call_id, send_json)

        def _call_requested_tool() -> PluginToolCallCompleteDict:
            assert _LMS_TOOL_CALL_ASYNC.get(None) is None
            _LMS_TOOL_CALL_SYNC.set(call_context)
            call_result = implementation(**kwds)
            return PluginToolCallComplete(
                session_id=self.session_id,
                call_id=call_id,
                result=call_result,
            ).to_dict()

        return asyncio.to_thread(_call_requested_tool)

    async def _call_tool_implementation(
        self, tool_call: ProvideToolsCallTool, send_json: SendMessageAsync
    ) -> PluginToolCallCompleteDict:
        # Find tool implementation
        tool_name = tool_call.tool_name
        tool_details = self._provided_tools.get(tool_name, None)
        if tool_details is None:
            raise ServerRequestError(
                f"Plugin does not provide a tool named {tool_name!r}."
            )
        # Validate parameters against their specification
        params_struct, tool_impl, is_async = tool_details
        raw_kwds = tool_call.parameters
        try:
            parsed_kwds = convert(raw_kwds, params_struct)
        except Exception as exc:
            err_msg = f"Failed to parse arguments for tool {tool_name}: {exc}"
            raise ServerRequestError(err_msg)
        kwds = to_builtins(parsed_kwds)
        if is_async:
            return await self._call_async_tool(
                tool_call.call_id, tool_impl, kwds, send_json
            )
        return await self._call_sync_tool(tool_call.call_id, tool_impl, kwds, send_json)

    # TODO: Reduce code duplication with the ChatResponseEndpoint definition
    async def _call_tool(
        self, tool_call: ProvideToolsCallTool, send_json: SendMessageAsync
    ) -> None:
        call_id = tool_call.call_id
        abort_events = self._abort_events
        if call_id in abort_events:
            err_msg = f"Tool call already in progress for {call_id} in session {self.session_id}"
            raise ServerRequestError(err_msg)
        abort_events[call_id] = abort_event = asyncio.Event()
        logger = new_logger(__name__)
        logger.update_context(self._logger.event_context, call_id=call_id)
        try:
            async with create_task_group() as tg:
                tg.start_soon(
                    self._cancel_on_event,
                    tg,
                    abort_event,
                    f"Aborting tool_call {call_id}",
                )
                logger.info(f"Running tool call {call_id}")
                tool_call_response: PluginToolCallCompleteDict | PluginToolCallErrorDict
                try:
                    tool_call_response = await self._call_tool_implementation(
                        tool_call, send_json
                    )
                except Exception as exc:
                    # Only catch regular exceptions,
                    # allowing the server to time out for client process termination events
                    err_msg = "Error calling tool implementation"
                    logger.error(err_msg, exc_info=True, exc=repr(exc))
                    # TODO: Determine if it's worth sending the stack trace to the server
                    tool_name = tool_call.tool_name
                    ui_cause = f"{type(exc).__name__}: {exc}"
                    # Tool calling UI only displays the title, so also embed the cause directly
                    error_title = f"Error calling tool {tool_name} in plugin {self.plugin_name!r} ({ui_cause})"
                    error_details = SerializedLMSExtendedErrorDict(
                        title=error_title,
                        rootTitle=error_title,
                        cause=ui_cause,
                        stack="\n".join(format_tb(exc.__traceback__)),
                    )
                    tool_call_response = PluginToolCallErrorDict(
                        type="toolCallError",
                        sessionId=self.session_id,
                        callId=call_id,
                        error=error_details,
                    )
                await send_json(tool_call_response)
                tg.cancel_scope.cancel()
        finally:
            self._abort_events.pop(call_id, None)

    def abort_tool_call(self, call_id: str) -> None:
        abort_event = self._abort_events.get(call_id)
        if abort_event is not None:
            abort_event.set()
        # Any server notification will be sent from the tool calling task

    def _abort_all_calls(self) -> None:
        for abort_event in self._abort_events.values():
            abort_event.set()
        # Any server notifications will be sent from the tool calling tasks

    async def discard_session(self) -> None:
        await self._queue.put(None)

    async def receive_tool_calls(self, send_message: SendMessageAsync) -> None:
        session_queue = self._queue
        try:
            while True:
                tool_call = await session_queue.get()
                if tool_call is None:
                    break
                await self._call_tool(tool_call, send_message)
        finally:
            self._abort_all_calls()


# TODO: Define a common "PluginHookHandler" base class
@dataclass()
class ToolsProvider(Generic[TPluginConfigSchema, TGlobalConfigSchema]):
    """Handle accepting tools provider session requests."""

    plugin_name: str
    hook_impl: ToolsProviderHook
    plugin_config_schema: type[TPluginConfigSchema]
    global_config_schema: type[TGlobalConfigSchema]

    def __post_init__(self) -> None:
        self._logger = logger = new_logger(__name__)
        logger.update_context(plugin_name=self.plugin_name)
        self._call_handlers: dict[str, ToolCallHandler] = {}

    async def process_requests(
        self, ws_session: _AsyncSessionPlugins, notify_ready: Callable[[], Any]
    ) -> None:
        """Create plugin channel and wait for server requests."""
        logger = self._logger
        endpoint = ToolsProviderEndpoint()
        # Async API expects timeouts to be handled via task groups,
        # so there's no default timeout to override when creating the channel
        async with ws_session._create_channel(endpoint) as channel:
            notify_ready()
            logger.info("Opened channel to receive tools session requests...")
            send_message = channel.send_message
            async with create_task_group() as tg:
                logger.debug("Waiting for tools session requests...")
                async for contents in channel.rx_stream():
                    logger.debug(f"Handling tools provider channel message: {contents}")
                    for event in endpoint.iter_message_events(contents):
                        logger.debug("Handling tools provider channel event")
                        endpoint.handle_rx_event(event)
                        match event:
                            case ProvideToolsDiscardSessionEvent():
                                await self._discard_session(event.arg)
                            case ProvideToolsInitSessionEvent():
                                logger.debug("Running tools listing hook")
                                ctl = ToolsProviderController(
                                    ws_session,
                                    event.arg,
                                    self.plugin_config_schema,
                                    self.global_config_schema,
                                )
                                tg.start_soon(self._invoke_hook, ctl, send_message)
                            case ProvideToolsCallToolEvent(_):
                                tg.start_soon(self._call_tool, event.arg)
                            case ProvideToolsAbortCallEvent(_):
                                self._abort_tool_call(event.arg)
                            case ChannelFinishedEvent(_):
                                pass
                            case _:
                                assert_never(event)
                    if endpoint.is_finished:
                        break

    async def _discard_session(self, session_id: str) -> None:
        """Abort the specified tools session (if it is still running)."""
        call_handler = self._call_handlers.get(session_id, None)
        if call_handler is not None:
            await call_handler.discard_session()

    async def _call_tool(self, tool_call_request: ProvideToolsCallTool) -> None:
        """Call the specified tool."""
        call_handler = self._call_handlers.get(tool_call_request.session_id, None)
        if call_handler is not None:
            await call_handler.start_tool_call(tool_call_request)

    def _abort_tool_call(self, abort_request: ProvideToolsAbortCall) -> None:
        """Abort the specified tool call (if it is still running)."""
        call_handler = self._call_handlers.get(abort_request.session_id, None)
        if call_handler is not None:
            call_handler.abort_tool_call(abort_request.call_id)

    async def _run_tools_session(
        self,
        session_id: str,
        provided_tools: ClientToolMap,
        send_json: SendMessageAsync,
    ) -> None:
        logger = self._logger
        call_handlers = self._call_handlers
        if session_id in call_handlers:
            err_msg = f"Tools session already in progress for {session_id}"
            raise ServerRequestError(err_msg)
        call_handler = call_handlers[session_id] = ToolCallHandler(
            self.plugin_name, session_id, provided_tools, self._logger.event_context
        )
        try:
            logger.info(f"Running tools session {session_id}")
            await call_handler.receive_tool_calls(send_json)
        finally:
            call_handlers.pop(session_id, None)
        logger.info(f"Terminated tools session {session_id}")

    async def _invoke_hook(
        self,
        ctl: ToolsProviderController[TPluginConfigSchema, TGlobalConfigSchema],
        send_json: SendMessageAsync,
    ) -> None:
        logger = self._logger
        session_id = ctl.session_id
        error_details: SerializedLMSExtendedErrorDict | None = None
        try:
            plugin_tools_list = await self.hook_impl(ctl)
            llm_tools_array, provided_tools = ChatResponseEndpoint.parse_tools(
                plugin_tools_list,
                allow_async=True,
            )
            llm_tools_list = llm_tools_array.to_dict()["tools"]
            assert llm_tools_list is not None  # Ensured by the parse_tools method
        except Exception as exc:
            err_msg = "Error calling tools listing hook"
            logger.error(err_msg, exc_info=True, exc=repr(exc))
            # TODO: Determine if it's worth sending the stack trace to the server
            error_title = f"Tools listing error in plugin {self.plugin_name!r}"
            ui_cause = f"{err_msg}\n({type(exc).__name__}: {exc})"
            error_details = SerializedLMSExtendedErrorDict(
                title=error_title,
                rootTitle=error_title,
                cause=ui_cause,
                stack="\n".join(format_tb(exc.__traceback__)),
            )
            error_message = ProvideToolsInitFailedDict(
                type="sessionInitializationFailed",
                sessionId=session_id,
                error=error_details,
            )
            await send_json(error_message)
            return
        init_message = ProvideToolsInitializedDict(
            type="sessionInitialized",
            sessionId=session_id,
            toolDefinitions=llm_tools_list,
        )
        await send_json(init_message)
        # Wait for further messages (until the session is discarded)
        await self._run_tools_session(session_id, provided_tools, send_json)


async def run_tools_provider(
    plugin_name: str,
    hook_impl: ToolsProviderHook,
    plugin_config_schema: type[BaseConfigSchema],
    global_config_schema: type[BaseConfigSchema],
    session: _AsyncSessionPlugins,
    notify_ready: Callable[[], Any],
) -> None:
    """Accept tools provider session requests."""
    tools_provider = ToolsProvider(
        plugin_name, hook_impl, plugin_config_schema, global_config_schema
    )
    await tools_provider.process_requests(session, notify_ready)



# Plugin runner doesn't actually care these type hints are incorrect
# It *does* care that an exception is raised
class ExampleCustomException(Exception):
    pass


async def preprocess_prompt(_ctl: None, _message: None) -> dict[None, None]:
    raise ExampleCustomException("Example plugin hook failure")



#####################################################################
# This files has been automatically generated from:
#   ../async/test_model_handles_async.py
#
# DO NOT EDIT THIS FILE! Edit the async test case listed above,
# and regenerate the synchronous test cases with async2sync.py
#####################################################################
"""Test model handles (LLM, EmbeddingModel) with the API.

Because these methods are defined specifically such that ALL THEY DO
is pass in a `self.identifier` parameter to the parent functions,
any errors in here are likely to be in the parent functions.
"""

import logging

import pytest
from pytest import LogCaptureFixture as LogCap

from lmstudio import Client, PredictionResult

from ..support import (
    EXPECTED_EMBEDDING,
    EXPECTED_EMBEDDING_ID,
    EXPECTED_EMBEDDING_LENGTH,
    EXPECTED_LLM,
    EXPECTED_LLM_ID,
    SHORT_PREDICTION_CONFIG,
)

# TODO: include several mock-based test cases here, as the important
#       functionality is actually in ensuring that the model wrappers
#       call the underlying session APIs with the model identifier
#       filled in appropriately. These cases can also then be executed
#       in CI without needing a live LM Studio instance.


@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_LLM, EXPECTED_LLM_ID))
def test_completion_llm_handle_sync(model_id: str, caplog: LogCap) -> None:
    prompt = "Hello"

    caplog.set_level(logging.DEBUG)
    with Client() as client:
        session = client.llm
        lm = session.model(model_id)
        response = lm.complete(prompt=prompt, config=SHORT_PREDICTION_CONFIG)
    # The continuation from the LLM will change, but it won't be an empty string
    logging.info(f"LLM response: {response!r}")
    assert isinstance(response, PredictionResult)
    assert response.content


@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_EMBEDDING, EXPECTED_EMBEDDING_ID))
def test_embedding_handle_sync(model_id: str, caplog: LogCap) -> None:
    text = "Hello, world!"

    caplog.set_level(logging.DEBUG)
    with Client() as client:
        session = client.embedding
        embedding = session.model(model_id)
        response = embedding.embed(input=text)
    logging.info(f"Embedding response: {response}")
    assert response
    assert isinstance(response, list)
    assert len(response) == EXPECTED_EMBEDDING_LENGTH



#####################################################################
# This files has been automatically generated from:
#   ../async/test_sdk_bypass_async.py
#
# DO NOT EDIT THIS FILE! Edit the async test case listed above,
# and regenerate the synchronous test cases with async2sync.py
#####################################################################
"""Test bypassing the SDK and accessing the API server directly."""

# This test case helps to differentiate between actual SDK failures
# and general failures due to an incorrect test environment setup.

import logging
import uuid
import warnings

from typing import Any, ContextManager

import pytest

from httpx_ws import connect_ws, WebSocketSession

from lmstudio import Client


@pytest.mark.lmstudio
def test_connect_and_predict_sync(caplog: Any) -> None:
    # Access the default API host directly
    api_host = Client.find_default_local_api_host()
    model_identifier = "hugging-quants/llama-3.2-1b-instruct"
    prompt = "Hello"

    caplog.set_level(logging.DEBUG)
    ws_cm: ContextManager[WebSocketSession] = connect_ws(f"ws://{api_host}/llm")

    with ws_cm as ws:
        # Authenticate
        client_identifier = str(uuid.uuid4())
        client_passkey = str(uuid.uuid4())
        auth_message = {
            "authVersion": 1,
            "clientIdentifier": client_identifier,
            "clientPasskey": client_passkey,
        }
        ws.send_json(auth_message)
        auth_result = ws.receive_json()
        if not auth_result["success"]:
            raise Exception(f"Authentication failed: {auth_result['error']}")

        # Send prediction request
        channel_id = 1
        creation_parameter = {
            "modelSpecifier": {
                "type": "query",
                "query": {"identifier": model_identifier},
            },
            "history": {
                "messages": [
                    {"role": "user", "content": [{"type": "text", "text": prompt}]}
                ]
            },
            "predictionConfigStack": {"layers": []},
        }
        create_message = {
            "type": "channelCreate",
            "endpoint": "predict",
            "channelId": channel_id,
            "creationParameter": creation_parameter,
        }
        ws.send_json(create_message)

        # Process prediction response
        fragments = []
        while True:
            message = ws.receive_json()

            match message:
                case {"type": "channelSend", "message": contents}:
                    match contents:
                        case {"type": "fragment", "fragment": {"content": fragment}}:
                            fragments.append(fragment)
                        case {"type": "fragment", "fragment": fragment}:
                            # Older LM Studio instance where message fragments were just strings
                            fragments.append(fragment)
                        case {"type": "success"}:
                            # We're done here
                            break
                case {"type": "channelError", "error": error}:
                    raise Exception(f"Prediction error: {error}")
                case {"type": "communicationWarning", "warning": warning}:
                    warnings.warn(f"Channel warning: {warning}")

    response = "".join(fragments)

    # The initial query from the LLM will change, but we expect it to be a question
    logging.info(f"LLM response: {response!r}")
    assert "?" in response
